// RMRClient.cc - Remote Message Router Client side

/* Author:  Marty Kraimer Date:    08APR1998 */

/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
 
#include <semLib.h>
#include <taskLib.h>

#include "FreeList.h"
#include "DLList.h"
#include "Message.h"
#include "MessagePvt.h"
#include "ConnectMessage.h"
#include "Tcp.h"
#include "WatchDog.h"

static const int secondsBetweenRetry = 10;

enum connectionState {stateStart,stateWaitConnect,stateConnected};

static const char *connectionStateString[] =
{ "stateStart","stateWaitConnect","stateConnected" };

enum messageClass {messageClassBind, messageClassUser};

static void connectCallback(void *pvt,tcpConnectState state);
static void receiveCallback(void *pvt,void *buffer,int32 nbytes);
static void RMRsendTask(void *pvt);

enum bindState {
    bindStateBinding,bindStateNotBound,bindStateConnected,bindStateNoServer
};
struct ServerInfo {
    DLList 	*pClientInfoList;
    char	*server;
    void	*routerServerPvt;
    bindState	state;
};

struct ClientInfo {
    MessageClient *pMessageClient;
    ServerInfo	  *pServerInfo;
};

struct SendNode {
    Message	*pmessage;
    ClientInfo	*pClientInfo;
};

struct RMRconnectMessage {
    char	magic[4];		// will be MPF
    int32	determineByteOrder;	// sent as 1
};

class RMRClient: public MessageRouterClient
{
public:
    RMRClient(int location,Tcp *pTcp,int buffersize,int queueSize);
    virtual ~RMRClient() {}
    virtual void bind(const char *server,MessageClient *messageClient);
    virtual int send(Message *message);
    virtual void report();
    //Remaining is for private use by this source module
    void	lock() {semTake(lockSem,WAIT_FOREVER);}
    void	unlock() {semGive(lockSem);}
    void	connect();
    void	disconnect();
    void	receiveConnect(void *buffer);
    void 	bindCallClient(ClientInfo *,connectStatus);
    void	receive(void *buffer,int32 nbytes);
    Tcp		*pTcp;	
    const int	sendBufferSize;
    bool	sameByteOrder;
    connectionState state;
    bool	retryBindRequests;
    char	*sendBuffer;
    DLList	*pServerInfoList;
    DLList	*pbindList;
    DLList	*psendList;
    DLList	*pnotConnectedList;
    FreeList	*pFreeSendNode;
    SEM_ID	sendRequest;
    SEM_ID	lockSem; //for modifying any list;
    int		sendTaskid;
    WatchDog	*wdId;
    int		queueSize;
    int		sendQueueFull;
};

int tcpMessageRouterClientStart(int location,int port,char *address,
    int bufsize,int queueSize)
{
    if(bufsize<100) {
	printf("bufsize <100. Set = 1024\n");
	bufsize = 1024;
    }
    if(queueSize<=0) {
	printf("queueSize %d. Set = 100\n",queueSize);
	queueSize = 100;
    }
    Tcp *pTcp = new Tcp(false,port,address,bufsize);
    new RMRClient(location,pTcp,bufsize,queueSize);
    return(0);
}

static void retryBind(void *pvt)
{
    RMRClient *pRMRClient = (RMRClient *)pvt;
    semGive(pRMRClient->sendRequest);
    pRMRClient->retryBindRequests = true;
    pRMRClient->wdId->start(secondsBetweenRetry,retryBind,(void *)pRMRClient);
}

static char taskname[] = "RMRClient";

RMRClient::RMRClient(int location,Tcp *ptcp,int bufsize,int queueSize)
  : MessageRouterClient(location), pTcp(ptcp), sendBufferSize(bufsize),
    sameByteOrder(false), state(stateStart), retryBindRequests(false),
    queueSize(queueSize), sendQueueFull(0)
{
    sendBuffer = new char[sendBufferSize];
    pServerInfoList = new DLList;
    pbindList = new DLList;
    psendList = new DLList;
    pnotConnectedList = new DLList;
    pFreeSendNode = new FreeList((int)sizeof(SendNode),100);
    sendRequest = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
    lockSem = semBCreate(SEM_Q_FIFO, SEM_FULL);
    sendTaskid = taskSpawn(taskname,110,VX_FP_TASK,5000,
	(FUNCPTR)RMRsendTask, (int)this,0,0,0,0,0,0,0,0,0);
    if(sendTaskid==ERROR) {
       printf("RMRClient taskSpawn Failure\n");
    }
    wdId = new WatchDog;
    wdId->start(secondsBetweenRetry,retryBind,(void *)this);
    pTcp->connect((void *)this,connectCallback,receiveCallback);
    prouter->registerClientRouter(this,getLocation());
}

static void connectCallback(void *pvt,tcpConnectState state)
{
    RMRClient	*pRMRClient = (RMRClient *)pvt;
    switch(state) {
        case tcpConnectYes: pRMRClient->connect(); break;
        case tcpConnectNo: pRMRClient->disconnect(); break;
        default:
	    printf("MRMClient:connectCallback illegal state %d\n",(int)state);
    }
}

static void receiveCallback(void *pvt,void *buffer,int32 nbytes)
{
    RMRClient *pRMRClient = (RMRClient *)pvt;
    pRMRClient->lock();
    connectionState state = pRMRClient->state;
    pRMRClient->unlock();
    switch(state) {
    case stateWaitConnect: pRMRClient->receiveConnect(buffer); break;
    case stateConnected: pRMRClient->receive(buffer,nbytes); break;
    default:
	printf("tcpRouter: receiveCallback illegal state\n");
	break;
    }
}

void RMRClient::bind(
    const char *server,MessageClient *pMessageClient)
{
    ClientInfo *pClientInfo = new ClientInfo;
    pMessageClient->routerPvt = (void *)pClientInfo;
    pClientInfo->pMessageClient = pMessageClient;
    ServerInfo *pServerInfo = 0;
    //see if already using this server
    lock();
    DLNode *pserverInfoNode = pServerInfoList->getFirst();
    while(pserverInfoNode) {
	pServerInfo = (ServerInfo *)pserverInfoNode->obj;
	if(::strcmp(pServerInfo->server,server)==0) break;
	pserverInfoNode = pServerInfoList->getNext(pserverInfoNode);
    }
    bool newServer = (pserverInfoNode ? false : true);
    if(newServer) {
	pServerInfo = new ServerInfo;
	pServerInfo->pClientInfoList = new DLList;
	pServerInfo->server = new char[::strlen(server)+1];
	::strcpy(pServerInfo->server,server);
	pServerInfo->routerServerPvt = 0;
	pServerInfo->state = bindStateNotBound;
	// add to serverList
	pserverInfoNode = DLList::allocNode();
	pserverInfoNode->obj = (void *)pServerInfo;
	pServerInfoList->addEnd(pserverInfoNode);
    }
    //add ClientInfo to pClientInfoList
    pClientInfo->pServerInfo = pServerInfo;
    DLNode *pclientInfoNode = DLList::allocNode();
    pclientInfoNode->obj = (void *)pClientInfo;
    pServerInfo->pClientInfoList->addEnd(pclientInfoNode);
    if(newServer) {
	// add to bindList
	pserverInfoNode = DLList::allocNode();
	pserverInfoNode->obj = (void *)pServerInfo;
	pbindList->addEnd(pserverInfoNode);
	pServerInfo->state = bindStateBinding;
	unlock();
	// make sure bind request gets sent
	semGive(sendRequest);
    } else {
        unlock();
	if(pServerInfo->state==bindStateConnected)
	    bindCallClient(pClientInfo,connectYes);
    }
}

int RMRClient::send(Message *pmessage)
{
    MessagePvt *msgPvt = getMessagePvt(pmessage);
    if(!msgPvt) return(-1);
    MessageClient *pMessageClient = msgPvt->pMessageClient;
    ClientInfo *pClientInfo = (ClientInfo *)pMessageClient->routerPvt;
    ServerInfo *pServerInfo = pClientInfo->pServerInfo;
    if(pServerInfo->state!=bindStateConnected) return(-1);
    if(psendList->getCount() >= queueSize) {
	++sendQueueFull;
	return(-1);
    }
    lock();
    SendNode *pSendNode = (SendNode *)pFreeSendNode->calloc();
    pSendNode->pmessage = pmessage;
    pSendNode->pClientInfo = pClientInfo;
    DLNode *node = DLList::allocNode();
    node->obj = (void *)pSendNode;
    psendList->addEnd(node);
    unlock();
    semGive(sendRequest);
    return(0);
}

static const char *bindStateString[4] = {
"binding","notBound","connected","no server"
};
void RMRClient::report()
{
    ServerInfo	*pServerInfo = 0;
    printf(" %d RMRClient %s queueSize %d inQueue %d sendQueueFull %d\n",
	getLocation(),connectionStateString[state],
	queueSize,psendList->getCount(),
	sendQueueFull);
    DLNode *node = pServerInfoList->getFirst();
    while(node) {
	pServerInfo = (ServerInfo *)node->obj;
	printf("   Server %s has %d clients. bindState %s\n",
	    pServerInfo->server,
	    pServerInfo->pClientInfoList->getCount(),
	    bindStateString[pServerInfo->state]);
	node = pServerInfoList->getNext(node);
    }
    pTcp->printStats();
}

void RMRClient::connect()
{
    RMRconnectMessage message;
    
    ::strcpy(message.magic,"MPF");
    message.determineByteOrder = 1;
    lock();
    int32 status = pTcp->sendFlush(&message,sizeof(message));
    if(status) {
	printf("RMRClient::connect tcp status is %d\n",status);
	taskSuspend(0);
    }
    state = stateWaitConnect;
    unlock();
}

void RMRClient::disconnect()
{
printf("RMRClient::disconnect\n");
    lock();
    state = stateStart;
    DLNode *pserverInfoNode = pServerInfoList->getFirst();
    while(pserverInfoNode) {
	ServerInfo *pServerInfo = (ServerInfo *)pserverInfoNode->obj;
	if(pServerInfo->state==bindStateConnected) {
	    DLNode *pclientInfoNode = pServerInfo->pClientInfoList->getFirst();
	    while(pclientInfoNode) {
	        ClientInfo *pClientInfo = (ClientInfo *)pclientInfoNode->obj;
	        bindCallClient(pClientInfo,connectNo);
	        pclientInfoNode =
		    pServerInfo->pClientInfoList->getNext(pclientInfoNode);
	    }
	}
	// now add to notConnectedList
	if(pServerInfo->state!=bindStateNotBound) {
	    pServerInfo->state = bindStateNotBound;
	    DLNode *node = DLList::allocNode();
	    node->obj = (void *)pServerInfo;
	    pnotConnectedList->addEnd(node);
	}
	pserverInfoNode = pServerInfo->pClientInfoList->getNext(pserverInfoNode);
    }
    // empty bindList and sendList
    DLNode *node;
    while((node = pbindList->getFirst())) pbindList->freeNode(node);
    while((node = psendList->getFirst())) {
	SendNode *pSendNode = (SendNode *)node->obj;
	psendList->freeNode(node);
	Message *pmessage = pSendNode->pmessage;
	pFreeSendNode->free(pSendNode);
	delete pmessage;
    }
    unlock();
}

void RMRClient::receiveConnect(void *buffer)
{
    RMRconnectMessage *pmessage = (RMRconnectMessage *)buffer;
    if((::strncmp(pmessage->magic,"MPF",4) != 0)) {
        printf("RMRClient::receiveConnect received illegal header\n");
	taskSuspend(0);
    }
    if(pmessage->determineByteOrder==1) {
        sameByteOrder = true;
    } else {
        sameByteOrder = false;
    }
    state = stateConnected;
    //Following will issue bind requests if any are in list
    semGive(sendRequest);
}

void RMRClient::bindCallClient(ClientInfo *pClientInfo,connectStatus status)
{
    ServerInfo *pServerInfo = pClientInfo->pServerInfo;
    MessageClient *pMessageClient = pClientInfo->pMessageClient;
    ConnectMessage *pmessage = new ConnectMessage;
    pmessage->status = status;
    MessagePvt *msgPvt = getMessagePvt((Message *)pmessage);
    msgPvt->pMessageClient = pMessageClient;
    msgPvt->pMessageRouterClient = this;
    msgPvt->routerServerPvt = pServerInfo->routerServerPvt;
    pMessageClient->bindCallback(pmessage);
}

void RMRClient::receive(void *buffer,int32 totalBytes)
{
    const char *pnext = (char *)buffer;
    int32 nbytes = 0;
    int32 temp;
    if(sameByteOrder) nbytes += MPF::fromBuffer(&pnext,&temp);
    else nbytes += MPF::fromBufferSwitch(&pnext,&temp);
    messageClass msgclass = (messageClass)temp;
    switch(msgclass) {
        case messageClassBind: {
	    ServerInfo *pServerInfo;
	    //bind message return: pServerInfo,routerServerPvt
	    nbytes += MPF::fromBuffer(&pnext,(void *)&pServerInfo,
		(int32)sizeof(void *));
	    nbytes += MPF::fromBuffer(&pnext,&pServerInfo->routerServerPvt,
		(int32)sizeof(void *));
	    if(pServerInfo->routerServerPvt) {//did connect to server
		pServerInfo->state = bindStateConnected;
	        DLList *pClientInfoList = pServerInfo->pClientInfoList;
	        DLNode *node = pClientInfoList->getFirst();
	        while(node) {
		    ClientInfo *pClientInfo = (ClientInfo *)node->obj;
		    bindCallClient(pClientInfo,connectYes);
		    node = pClientInfoList->getNext(node);
	        }
	    } else { // server was probably not registered
		pServerInfo->state = bindStateNoServer;
		DLNode *node = DLList::allocNode();
		node->obj = (void *)pServerInfo;
		pnotConnectedList->addEnd(node);
	    }
	    break;
	}
        case messageClassUser: {
	    // client message: pClientInfo,message
	    ClientInfo *pClientInfo = 0;
	    nbytes += MPF::fromBuffer(&pnext,
		(void *)&pClientInfo,(int32)sizeof(pClientInfo));
            int32 mtype = 0;
            int32 nb = 0;
	    if(sameByteOrder)  nb = MPF::fromBuffer(&pnext,&mtype);
	    else nb = MPF::fromBufferSwitch(&pnext,&mtype);
	    pnext -= nb;
	    Message *pmessage = Message::allocate((messageType)mtype);
	    MessagePvt *msgPvt = getMessagePvt(pmessage);
	    msgPvt->pMessageClient = pClientInfo->pMessageClient;
	    msgPvt->pMessageRouterClient = this;
	    msgPvt->routerServerPvt = pClientInfo->pServerInfo->routerServerPvt;
	    if(sameByteOrder) nbytes += pmessage->fromBuffer(&pnext);
	    else nbytes += pmessage->fromBufferSwitch(&pnext);
	    MessageClient *pMessageClient = pClientInfo->pMessageClient;
	    pMessageClient->clientCall(pmessage,pMessageClient->getClientPvt());
	    break;
	}
        default:
	    printf("RMRClient::receive illegal message class %d\n",
		(int)msgclass);
	    taskSuspend(0);
    }
    if(nbytes!=totalBytes) {
	printf("RMRClient::receive received incorrect message size"
	    " received %d used %d\n",totalBytes,nbytes);
    }
}

static void RMRsendTask(void *pvt)
{
    RMRClient	*pRMRClient = (RMRClient *)pvt;
    while(true) {
	semTake(pRMRClient->sendRequest,WAIT_FOREVER);
	if(pRMRClient->state!=stateConnected) continue;
	// If time try all notBound requests
	if(pRMRClient->retryBindRequests) {
	    DLNode *node;
	    pRMRClient->lock();
	    while((node = pRMRClient->pnotConnectedList->getFirst())) {
		ServerInfo *pServerInfo = (ServerInfo *)node->obj;
		pRMRClient->pnotConnectedList->freeNode(node);
		pServerInfo->state = bindStateBinding;
		node = DLList::allocNode();
		node->obj = (void *)pServerInfo;
		pRMRClient->pbindList->addEnd(node);
	    }
	    pRMRClient->unlock();
	    pRMRClient->retryBindRequests = false;
	}
	//send all outstanding bind requests
	// bind request is: messageClassBind,server,pServerInfo
	while(true) {
	    ServerInfo *pServerInfo = 0;
	    pRMRClient->lock();
	    DLNode *node = pRMRClient->pbindList->getFirst();
	    if(node) {
		pServerInfo = (ServerInfo *)node->obj;
		pRMRClient->pbindList->freeNode(node);
	    } else {
		pServerInfo = 0;
	    }
	    pRMRClient->unlock();
	    if(pServerInfo==0) break;
	    int32 messageLength = 0;
	    char *buffer = pRMRClient->sendBuffer;
	    int32 temp = messageClassBind;
	    messageLength += MPF::toBuffer(&buffer,temp);
	    messageLength += MPF::toBuffer(&buffer,pServerInfo->server);
	    messageLength += MPF::toBuffer(&buffer,(void *)&pServerInfo,
		(int32)sizeof(void *));
	    pRMRClient->pTcp->send(pRMRClient->sendBuffer,messageLength);
	}
	//Now send all outstanding client send requests
	//Each message is messageClassUser,routerServerPvt,pClientInfo,message
	while(true) {
	    SendNode *pSendNode = 0;
	    pRMRClient->lock();
	    DLNode *node = pRMRClient->psendList->getFirst();
	    if(node) {
		pSendNode = (SendNode *)node->obj;
		pRMRClient->psendList->freeNode(node);
	    } else {
		pSendNode = 0;
	    }
	    pRMRClient->unlock();
	    if(pSendNode==0) break;
	    Message *pmessage = pSendNode->pmessage;
	    ClientInfo *pClientInfo = pSendNode->pClientInfo;
	    pRMRClient->pFreeSendNode->free(pSendNode);
	    ServerInfo *pServerInfo = pClientInfo->pServerInfo;
	    int32 messageLength = 0;
	    char *buffer = pRMRClient->sendBuffer;
	    int32 temp = messageClassUser;
	    messageLength += MPF::toBuffer(&buffer,temp);
	    messageLength += MPF::toBuffer(&buffer,
		&pServerInfo->routerServerPvt,(int32)sizeof(void *));
	    messageLength += MPF::toBuffer(&buffer,(void *)&pClientInfo,
		(int32)sizeof(void *));
	    messageLength += pmessage->toBuffer(&buffer);
	    pRMRClient->pTcp->send(pRMRClient->sendBuffer,messageLength);
	    delete pmessage;
	}
	pRMRClient->pTcp->flush();
    }
}
